/*---------------+---------------------------------------------------------------+--------------+
                 | Author: Daniel Porumbel 2018       daniel.porumbel@cnam.fr    |
                 |License: Creative Commons Attribution 3.0 Unported License     |
                 |         http://creativecommons.org/licenses/by/3.0/           |             
                 +--------------------------------------------------------------*/
#include <iostream>
#include <iomanip>
#include "bendersSubprob.h"
#include <list>
#include <vector>
#include <ilcplex/ilocplex.h>                    //Cplex header
using namespace std;

#define THREADS 1
#define PRINT_OPTSOL_OF_BENDERS_POLYTOPE

int isEdge(int v1, int v2, double * y)
{
    if(y==NULL)
        return g[v1][v2];
    if(g[v1][v2])
        if(y[gEdge[v1][v2]]>EPSILON)
            return 1;
    return 0;
}

void fillConnCompUpToDepth_BFS(int* connected, double* y, int source, int depth)
{
    list<int> queue;
    list<int> queueDepth;
    queue.push_back(source); 
    queueDepth.push_back(depth); 
    connected[source] = 1;

    while(!queue.empty()){
        int currentVtx   = queue.front();
        queue.pop_front();
        int depthCurrVtx = queueDepth.front();
        queueDepth.pop_front();
        if(depthCurrVtx>0)
          for(int v=0;v<n;v++)
            if(v!=currentVtx)
                if(isEdge(currentVtx,v,y))
                    if(!connected[v]){
                        connected[v] = 1;
                        queue.push_back(v);
                        queueDepth.push_back(depthCurrVtx-1);
                    }
    }
}
void fillConnCompUpToDepth(int* connected, double* y, int currentVtx, int depth)
{
    fillConnCompUpToDepth_BFS(connected,y,currentVtx,depth);
    for(int i=0;i<n;i++)
        if(connected[i]>0)
            connected[i] = 1;
}

int newCutSet(int* connected, double *cutEdges, double & bSinks)
{
    //int*sinkAdded = new int[n];
    //for(int v=0;v<n;v++)
    //    sinkAdded[v] = 0;
    bSinks = 0;
    for(int v=0;v<n;v++)
        if(!connected[v])
            bSinks+=b[v];
    for(int e=0;e<m;e++)
        cutEdges[e] = 0;
    for(int e=0;e<m;e++){
        int vtx1 = ii[e];
        int vtx2 = jj[e];
        if(connected[vtx2]){
            int tmp = vtx1;
            vtx1    = vtx2;
            vtx2    = tmp;
        }
        //if there is any connected, then vtx1 is connected
        if(connected[vtx1]){
            if(!connected[vtx2]){
                cutEdges[e] = 1;
                //if(!sinkAdded[vtx2]){
                //    sinkAdded[vtx2] = 1;
                //    bSinks += b[vtx2];
                //}
            }
        }
    }//end scan edges
    //delete[] sinkAdded;
    return (bSinks>0);
}

//Input yyy of size m
//Output u of size m
void newCut(double * yyy, double *u, double &rightHand)
{
    int i;
    IloEnv env;
    IloModel model(env);
    IloNumVarArray uu(env, m+n-1, 0, 10000000);
    
    IloExpr expr(env);
    for(i=0;i<m;i++)
        expr-=(yyy[i])*uu[i];
    for(i=m;i<m+n-1;i++){
        expr+=(b[i-m])*uu[i];
        //clog<<"b["<<i<<"-"<<m<<"]="<<b[i-m]<<endl;
    }
    IloObjective obj(env, expr, IloObjective::Maximize);
    model.add(obj);

    IloRangeArray constr(env);
    IloExpr e(env);
    for(i=0;i<m+n-1;i++)
        e += uu[i];
        //e += (i+1)*uu[i];
    constr.add(e==1);

    for(i=0;i<m;i++){
        int vtx1 = ii[i];
        int vtx2 = jj[i];
        //clog<<"vtx1="<<vtx1<<","; clog<<"vtx2="<<vtx2<<endl;
        if((vtx1!=n-1)&&(vtx2!=n-1)){
            IloExpr e1(env);
            e1+=uu[i]-uu[vtx1+m]+uu[vtx2+m];
            constr.add(e1>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e1.end();

            IloExpr e2(env);
            e2+=uu[i]-uu[vtx2+m]+uu[vtx1+m];
            constr.add(e2>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e2.end();
        }
        if(vtx1==n-1){
            IloExpr e1(env);
            e1+=uu[i]+uu[vtx2+m];
            constr.add(e1>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e1.end();

            IloExpr e2(env);
            e2+=uu[i]-uu[vtx2+m];
            constr.add(e2>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e2.end();
        }
        if(vtx2==n-1){
            IloExpr e1(env);
            e1+=uu[i]-uu[vtx1+m];
            constr.add(e1>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e1.end();

            IloExpr e2(env);
            e2+=uu[i]+uu[vtx1+m];
            constr.add(e2>=0);
            //model.add(constr[cAdded]);
            //cAdded ++;
            e2.end();
        }
    }
    model.add(constr);
    //Solve model
    IloCplex cplex (env);
    cplex.extract(model);
    cplex.setOut(env.getNullStream());
    cplex.setError(env.getNullStream());
    cplex.setParam(IloCplex::Threads, THREADS);
    cplex.setParam(IloCplex::NumericalEmphasis, true);
    cplex.solve();

    //Extract solution
    double objVal = cplex.getObjValue();
    clog<<"          Classical separation cut to be added with objVal="<<objVal<<endl;
    IloNumArray solution(env);
    cplex.getValues (solution,uu);
    int printed = 0;
    for(i=0;i<m;i++){
            u[i] = solution[i];
            #ifdef PRINT_OPTSOL_OF_BENDERS_POLYTOPE
            if(i==0){
                printed = 0;
                clog<<"          sol sep:";
            }
            if(u[i]>0){
                printed ++;
                if(printed<5)
                    clog<<"u["<<i<<"]="<<u[i]<<",";         
            }
            #endif
    }
    rightHand = 0;
    for(i=m;i<m+n-1;i++){
        #ifdef PRINT_OPTSOL_OF_BENDERS_POLYTOPE
            if(i==m)
                printed = 0;
            if(u[i]>0){
                printed ++;
                if(printed<7)
                    clog<<"u["<<i<<"]="<<solution[i]<<",";         
            }
        #endif
        rightHand += b[i-m]*solution[i];
    }
    clog<<endl;
    clog<<"          generated separation constraint has rightHand="<<rightHand<<endl;
    //cplex.exportModel("test.lp");

    //Clean all and return
    constr.end();
    uu.end();
    model.end();
    env.end();
}


#define checkHitPointFillNewCutNew_tStarWhenNeeded(xbase,d,u,rHand)       \
 do{                                                                      \
     double leftValue ;                                                   \
     double rHandSep;                                                     \
     static double * contact = new double[m];                             \
     static double * rowSep  = new double[m];                             \
     do{                                                                  \
       for(i=0;i<m;i++)                                                   \
           contact[i] = xbase[i]+tStar*tStar_scale*d[i];                  \
       double xval = 0, dval = 0, cval = 0;                               \
       for(i=0;i<m;i++){                                                  \
           xval += xbase[i];                                              \
           dval += d[i];                                                  \
           cval += contact[i];                                            \
       }                                                                  \
       double d_prod_u = 0;                                               \
       newCut(contact,rowSep, rHandSep);                                  \
       leftValue = 0;                                                     \
       for(i=0;i<m;i++){                                                  \
           leftValue+=rowSep[i] * contact[i];                             \
           d_prod_u +=rowSep[i] * d[i];                                   \
       }                                                                  \
                                                                          \
       clog<<"    x"<<xval<<" d"<<dval<<" contact val"<<cval              \
           <<"d_prod_u="<<d_prod_u<<endl<<"    leftValSep="               \
           <<leftValue<<">= ? >="<<rHandSep<<endl;                        \
       if(leftValue<rHandSep-EPSILON){                                    \
           if(tStar==tStarInitial){                                       \
               rHand = rHandSep;                                          \
               for(i=0;i<(::m);i++)                                       \
                   u[i] = rowSep[i];                                      \
           }                                                              \
           tStar = tStar*0.8;                                             \
           cerr<<"Hit point infeasible, I decrease tStar to"<<tStar<<endl;\
           /*exit(1);*/                                                   \
       }                                                                  \
     }while(leftValue<rHandSep-EPSILON);                                  \
  }while(false);

#define printFirstSol(solution)                                           \
  do{                                                                     \
      clog<<"Objval in projection (tStar)="<<tStar<<endl;                 \
      clog<<"non normalized sol before minimizing coefs:";                \
      printed = 0;                                                        \
      for(i=0;i<m+n-1;i++)                                                \
          if(solution[i]>0) {                                             \
              clog<<solution[i]<<"/"<<i<<" ";                             \
              printed++;                                                  \
              if(printed==7)                                              \
                  break;                                                  \
          }                                                               \
      clog<<endl;                                                         \
  }while(false);

#define printObjFunc(xbase,d,b)                                           \
  do{                                                                     \
    for(i=0;i<m+n-1;i++){                                                 \
        if(i==0)                                                          \
            clog<<"      projection: objective func is (in vars uuu)";    \
        if((i==m)||(i==0))                                                \
            printed = 0;                                                  \
        if(printed<5){                                                    \
            if(i<m)                                                       \
                clog<<"x"<<i<<"="<<xbase[i]<<"(d"<<d[i]<<") ";            \
            else                                                          \
                clog<<"-b"<<i-m<<"="<<b[i-m]<<" ";                        \
        }                                                                 \
        printed ++;                                                       \
    }                                                                     \
    clog<<endl;                                                           \
  }while(false);

#define printSol(solution,xbase,d,b)                                      \
  do{                                                                     \
    for(i=0;i<m+n-1;i++){                                                 \
        if(i==0){                                                         \
            clog<<"Solution uuu minimizing coefs ";                       \
            printed = 0;                                                  \
        }                                                                 \
        if(i==m){                                                         \
            clog<<"   | |vtx:   ";                                        \
            printed = 0;                                                  \
        }                                                                 \
        if(solution[i]!=0){                                               \
            printed++;                                                    \
            if(printed<5){                                                \
                clog<<"u["<<i<<"]="<<solution[i]<<"/";                    \
                if(i<m){                                                  \
                    clog<<"x["<<i<<"]="<<xbase[i]<<":";                   \
                    clog<<"d["<<i<<"]="<<d[i]<<",";                       \
                }else                                                     \
                    clog<<"b["<<i-m<<"]="<<b[i-m]<<",";                   \
            }                                                             \
        }                                                                 \
    }                                                                     \
  }while(false);

#define printContribs(solution,xbase,b)                                   \
  do{                                                                     \
    double contribBase = 0;                                               \
    double contribRight = 0;                                              \
    for(i=0;i<m+n-1;i++){                                                 \
        if(i<m){                                                          \
            contribBase+=xbase[i]*solution[i];                            \
        }else{                                                            \
            contribRight-=(b[i-m])*solution[i];                           \
        }                                                                 \
    }                                                                     \
    clog<<"\n contribObjEdge, contribObjVtx="                             \
        <<contribBase<<","<<contribRight                                  \
        <<"cObjVtx+tStar="<<-contribRight+tStar/tStar_scale<<endl;        \
  }while(false);                                                          

//The projection sub-problem can have quite strong numerical problems because 
//u is unbounded and large values of u can be multiplied with close-to-zero 
//factors to generate non-zero values, not good. To get lower u, we use
//1) limit the max u value
//2) the Charness-Cooper constraint (denominator*u = 1) ~ (-\sum d_iju_ij = 1) 
//   replaces 1 by 1/tStar_scale => smaller u
//These techniques exclude some large u, but no problem: the final sol can
//be checked for feasibility using the primal constraints. Usually, it is ok!
#define MAX_U_VAL   100
double tStar_scale  = 1;
double newProjection(double* xbase,double*d,double*u, double&rHand, int checkHitPoint)
{
    clog<<"------bendersSubprob Projection (Intersect) subprob"<<endl;
    double   tStar = -1;
    int      i;
    IloEnv   env;
    IloModel model(env);
    IloNumVarArray uuu(env, m+n-1, 0, MAX_U_VAL);//first m vars for edges, 
                                                 //next n-1 for vertices excluding source
    IloNumArray solution(env);
    
    //Set objective function
    for(i=0;i<m;i++){
        if(xbase[i]<EPSILON)
            xbase[i] = 0;
    }
    IloExpr expr(env);
    for(i=0;i<m;i++)
        expr+=xbase[i]*uuu[i];
    for(i=m;i<m+n-1;i++)
        expr-=(b[i-m])*uuu[i];
    IloObjective obj(env, expr, IloObjective::Minimize);
    model.add(obj);
    #ifdef PRINT_OPTSOL_OF_BENDERS_POLYTOPE
    int printed = 0;
    printObjFunc(xbase,d,b);
    #endif

    //Set constraints
    IloRangeArray constr(env);
    IloExpr e(env);
    for(i=0;i<m;i++)
        if((d[i]>EPSILON)||(d[i]<-EPSILON))
            e -= d[i]*uuu[i];
    constr.add(e==1.0/tStar_scale);

    for(i=0;i<m;i++){
        int vtx1 = ii[i];
        int vtx2 = jj[i];
        //clog<<"vtx1="<<vtx1<<","; clog<<"vtx2="<<vtx2<<endl;
        if((vtx1!=n-1)&&(vtx2!=n-1)){
            IloExpr e1(env);
            e1+=uuu[i]-uuu[vtx1+m]+uuu[vtx2+m];
            constr.add(e1>=0);
            e1.end();

            IloExpr e2(env);
            e2+=uuu[i]-uuu[vtx2+m]+uuu[vtx1+m];
            constr.add(e2>=0);
            e2.end();
        }
        if(vtx1==n-1){
            IloExpr e1(env);
            e1+=uuu[i]+uuu[vtx2+m];
            constr.add(e1>=0);
            e1.end();

            IloExpr e2(env);
            e2+=uuu[i]-uuu[vtx2+m];
            constr.add(e2>=0);
            e2.end();
        }
        if(vtx2==n-1){
            IloExpr e1(env);
            e1+=uuu[i]-uuu[vtx1+m];
            constr.add(e1>=0);
            e1.end();

            IloExpr e2(env);
            e2+=uuu[i]+uuu[vtx1+m];
            constr.add(e2>=0);
            e2.end();
        }
    }
    model.add(constr);
    IloCplex cplex (env);
    cplex.extract(model);
    cplex.setOut(env.getNullStream());
    cplex.setError(env.getNullStream());
    cplex.setWarning(env.getNullStream());
    cplex.setParam(IloCplex::Threads, THREADS);
    cplex.setParam(IloCplex::NumericalEmphasis, true);

    //Solve model
    try{
        cplex.solve();
        tStar = cplex.getObjValue();
        cplex.getValues (solution,uuu);
        #ifdef PRINT_OPTSOL_OF_BENDERS_POLYTOPE
        printFirstSol(solution);
        #endif
        if(tStar<-EPSILON){
            //this can be real error or numerical error, check with std separ
            static double * rowSepInf  = new double[m];
            double rHndInf;
            newCut(xbase,rowSepInf,rHndInf);
            double leftValue = 0;
            for(i=0;i<m;i++)
                leftValue+=rowSepInf[i] * xbase[i];
            clog<<"leftValSepp="<<leftValue<<">= ? >="<<rHndInf<<endl;
            if(leftValue<rHndInf-EPSILON){
                cerr<<"Error: input xbase seems infeasible="<<leftValue<<">= ? >="<<rHndInf<<endl;
                exit(1);
            }
        }
        if(tStar<0)
            tStar = 0;

        if(checkHitPoint){
            double tStarInitial = tStar;
            checkHitPointFillNewCutNew_tStarWhenNeeded(xbase,d,u,rHand);
            if(tStar!=tStarInitial)
                return tStar * tStar_scale;        //I think tStar_scale=1
        }

        //Search for constraint with lower coefs but same tStar obj val
        model.add(expr==tStar);
        for(i=0;i<m+n-1;i++)
            obj.setLinearCoef(uuu[i],1);
        //cplex.extract(model);   //This re-loads whole model, optimizing from scratch
        cplex.setParam(IloCplex::EpRHS, 1e-9);
        cplex.solve();
        cplex.getValues (solution,uuu);
        tStar*=tStar_scale;
    }catch(IloCplex::Exception e){
        cplex.exportModel("bendersProjectionPb.lp");
        cerr<<"\n\n\nError in projection subproblem"<<e.getMessage()
            <<"\nModel saved to bendersProjectionPb.lp.\n";
        //1256  CPXERR_SINGULAR //1217  CPXERR_NO_SOLN
        if(e.getStatus()==CPXERR_SINGULAR)
            cerr<<"\n I won't exit for Singular matrix error\n"
                  "this only happens for 2nd solve. I take first sol\n";
        else
            exit(1);
    }

    //Correct numerical problems of solution and calculate rHand
    for(i=0;i<m+n-1;i++){
        if(solution[i]<0)      //The cplex tolerance to accepting smthg>=0
            clog<<"Z";         //can cause some numerical probls
        if(solution[i]<EPSILON)//this can only be a numerical error
            solution[i] = 0;   //because the Benders polytope has u>=0
        if(i<m)
            u[i] = solution[i];
    }
    rHand = 0;
    for(i=m;i<m+n-1;i++)
        rHand += b[i-m]*solution[i];

    #ifdef PRINT_OPTSOL_OF_BENDERS_POLYTOPE
    printSol(solution,xbase,d,b);
    printContribs(solution,xbase,b);
    #endif

    clog<<"------ End Projection algo: found constraint with rightHand="<<rHand<<endl;

    //Clean all and return
    expr.end();
    constr.end();
    uuu.end();
    model.end();
    env.end();
    
    return tStar;
}
double newProjection(double* xbase,double*d,double*u, double&rHand){
    return newProjection(xbase,d,u,rHand,0);//do no check hit point
}
