/*---------------+---------------------------------------------------------------+--------------+
                 | Author: Daniel Porumbel 2018       daniel.porumbel@cnam.fr    |
                 |License: Creative Commons Attribution 3.0 Unported License     |
                 |         http://creativecommons.org/licenses/by/3.0/           |             
                 +--------------------------------------------------------------*/
#include "iographs.h"
#include "bendersSubprob.h"
#include <iostream>
#include <cmath>
#include <fstream>
#include <cstdlib>
#include <cassert>
#include <string.h>
#include <stdio.h>
using namespace std;

int  graphClass;
char graphClassName[20];    //an id of max 20 letters
int  minId[KNOWNCLASSES]                 ={INT_MIN  ,11        ,71        ,101      ,501      ,601      ,801      };
int  maxId[KNOWNCLASSES]                 ={   10    ,70        ,100       ,500      ,600      ,800      ,1200      };
char allGraphClassNames[KNOWNCLASSES][20]={"randomx","randomy","LAyered"  ,"layered","LAYered","nonunif","layer2y"};
double bMaxDemand[KNOWNCLASSES]          ={10       ,-2        ,10        ,10       ,50       ,50       ,  -2     };
                                         //random,^      ^           ^         ^       ^       ^             ^
                                         //              |           |         |       |       |             |
                                         //random with demand max 2, |         |       |       |             |
                                         //layered 10 by 10 with demand max 10 |       |       |             |
                                         //layered 20 by 20 with demand max 10-*       |       |             |
                                         //layered 10 by 10 with demand max 50---------*       |             |
                                         //nonuniform degrees----------------------------------*             | 
                                         //layered 20 by 20 with demand max 2 (min demand=1, see minus sign)-*
double bMaxDemandFromCmdLine=-1;         //using option -dABC changes (override) the max demand to integer ABC
int  idThisInstance;
char instanceName[100];
char* labelGraph = NULL;
double bandwidth  ;

void initDemands(int divFactor,int no_vtx_with_zero_demand){
    divFactor++;
    clog<<"Generating demands (vect b)\n";
    #ifdef FEWERSINKS
    divFact = 5;
    #endif
    b = new double[n];
    for(int i=0;i<n;i++){
            b[i] = (rand()%divFactor);        
    }
    if(no_vtx_with_zero_demand)
        for(int i=0;i<n-1;i++)
            if(b[i]==0)
                b[i] = 1;
    b[n-1] = 0;                 //This is the source
}
void initDemands(int maxDem){
    if(maxDem>0)
        initDemands(maxDem,0);
    else
        initDemands(-maxDem,1);       //no vtx without a demand
}

void initDataStructsLoadGraph(){
    g = new int*[n];
    gEdge = new int*[n];
    ii= new int[m];
    jj= new int[m];
    for(int i=0;i<n;i++){
        g[i] = new int[n];
        gEdge[i] = new int[n];
        for(int j=0;j<n;j++)
            gEdge[i][j] = -1;//no edge here by default
    }
    for(int i=0;i<n;i++)
        for(int j=0;j<n;j++)
            g[i][j] = 0;
}
void handleBandwidth(){
    clog<<"Attention: I consider a bandwidth of "<<bandwidth<<endl;
    for(int i=0;i<n;i++)
        b[i] = b[i]/bandwidth;
    clog<<"Here are the adapted equivalent demands:";
    for(int i=0;i<n;i++)
        clog<<b[i]<<",";
    clog<<endl;
}
int getRndBin(){
    int binSelector = rand()%18;
    if(binSelector<=2)
        return 3;
    if(binSelector<=6)
        return 2;
    if(binSelector<=11)
        return 1;
    if(binSelector<=17)
        return 0;
    cerr<<"Strange to get here. There is a problem in getRndBin() in iographs.cpp. Could not generate graph. Exit.\n";
    exit(1);
    return INT_MAX;
}
int rouletteWheel(){
    if(n%4!=0){
        cerr<<"I want a multiple of 4 for this graph class with non uniform degree distribution, because I'll need four classes.\n"; 
        exit(2);
    }
    int bin = getRndBin();
    return bin*(n/4)+rand()%(n/4);
}
int repairDegree(int v, int* degrees){
    int* connectComp = new int[n];
    for(int i=0;i<n;i++)
        connectComp[i] = 1;
    connectComp[n-1] = 1;
    //you can use below to make a better repair. Kept last version for compatibility with old version graph generators
    //fillConnectedUpToDepth(connectComp,NULL,n-1,n);
    clog<<"Trying to repair node of zero degree:..."<<v;
    for(int r=v+1;r<n-1;r++){
         if(degrees[r]>1){
              for(int e=0;e<m;e++){
                   if((ii[e]==r)||(jj[e]==r)){
                        int otherEnd;
                        if(ii[e]==r){
                            ii[e]=v;
                            otherEnd=jj[e];
                        }else{
                            jj[e]=v;
                            otherEnd=ii[e];
                        }
			if(connectComp[otherEnd]){
				clog<<"I will link the following edge at "<<v<<": "<<e<<"...";
				clog<<"Connecting "<<v<<" to "<<otherEnd<<"..";
				g[r][otherEnd]=0;g[otherEnd][r]=0;
				g[v][otherEnd]=1;g[otherEnd][v]=1;
				gEdge[r][otherEnd]=-1;gEdge[otherEnd][r]=-1;
				gEdge[v][otherEnd]=e; gEdge[otherEnd][v]=e;
				degrees[v]++;
				degrees[r]--;
				clog<<"Finished successfully.\n";
				delete[] connectComp;
				return 1;
			}
                   }
              }
         }
    }
    delete[] connectComp;
    clog<<"Sorry, I could not solve this\n";
    cerr<<"Sorry, I could not solve the zero degree of vertex"<<v<<endl;
    return 0;
}
void connectDisconnectedVtxsIfNeeded(){
    int* degrees = new int[n];
    for(int v=0;v<n;v++)
        degrees[v]=0;
    for(int e=0;e<m;e++){
        degrees[ii[e]]++;
        degrees[jj[e]]++;
    }
    for(int v=0;v<n;v++){
        if(degrees[v]==0)
            repairDegree(v,degrees);
    }
    delete[] degrees;
}
void loadGraph(){
    initDataStructsLoadGraph();
    int edgeId = 0;
    clog<<"Generating edges: ";
    for(;edgeId<m;edgeId++){
        clog<<edgeId<<","<<flush;
        int i,j;
        int triesEdgeAdditions = 0;
        do{
            int nn = n;
            //if(graphClass>0) nn = n+5;
            i = rand()%nn;
            if(graphClass<=1)         //pure random
                j = rand()%nn;
            else{
                j = 0;
                if((graphClass==3)||(graphClass==6))     
                    j = i+rand()%20;  //layered 20 by 20
                if(graphClass==2)
                    j = i+ rand()%10;
                if(graphClass==4)
                    j = i+ rand()%10;
                if(j>=n)
                    j=i;
            }
            if(graphClass==5){
                i = rouletteWheel();
                j = rouletteWheel();
            }
            if(i>=n) i = n- 1;
            if(j>=n) j = n- 1;
            triesEdgeAdditions++;
            if(triesEdgeAdditions>2*n){
                cerr<<"\nError:I give up on trying to generate a graph of the given class\n\
                       with "<<n<<" vertices and "<<m<<" edges. Did you think that\n\
                       a complete graph of this order has "<<n*(n-1)/2<<" edges?\n\
                       I can NOT continue.\n";
                exit(1);
            }
        }while((g[i][j]==1)||(i==j));
        g[i][j] = 1;
        g[j][i] = 1;
        ii[edgeId] = i; 
        jj[edgeId] = j;
        gEdge[i][j] = edgeId;
        gEdge[j][i] = edgeId;
        //clog<<"edgeId="<<edgeId<<"{"<<i<<","<<j<<"}"<<endl;
    }
    clog<<endl;
    connectDisconnectedVtxsIfNeeded();
}
int getClass(int idGraph){
    for(int i=0;i<KNOWNCLASSES;i++)
        if((idGraph>=minId[i])&&(idGraph<=maxId[i]))
            return i;
    cerr<<"Attention: graph class undefined for id="<<idGraph<<"\n";
    exit(1);
    return INT_MAX;
}
void setInstanceName(char* instName){
    instName[0] = 0;
    char padding[10];
    strcpy(padding,"");
    double bMaxDem = bMaxDemand[graphClass];
    if(bMaxDem<100)
        strcpy(padding,"0");
    if(bMaxDem<10)
        strcpy(padding,"00");
    char bufferInstName [200];
    sprintf(bufferInstName,"%s-demMax%s%g-edg%d-vtx%d-id%d",graphClassName,padding,bMaxDem,m,n,
                                (minId[graphClass]==INT_MIN?idThisInstance:idThisInstance-minId[graphClass]+1));
    if(bandwidth>1)
        sprintf(instName,"%s-bandwd%d",bufferInstName,(int)bandwidth);
    else
        sprintf(instName,"%s",bufferInstName);
}
void saveGraphAndDemands(char* filename){
    clog<<"Saving graph to "<<filename<<"....";
    clog.flush();
    ofstream fGraph;
    fGraph.open(filename,fstream::out);
    if(fGraph.fail()){
        cerr<<"Failed to open:"<<filename<<" in saveGraphAndDemands(..). Does the folder (./instances) exist?" <<endl;
        return;
    }
    fGraph<<"c SOURCE: Daniel Porumbel "<<endl;
    fGraph<<"c DESCRIPTION: testing ray projections in a Benders decomposition context."<<endl;
    if(graphClass==0){
        fGraph<<"c Graph Class: random graph with random demands between 0 and 10\n";
    }
    if(graphClass==1){
        fGraph<<"c Graph Class: random graph with random demands between 0 and 1\n";
    }
    if(graphClass==2){
        fGraph<<"c Layered Graph 10: all edges (vtx1,vtx2) satisfy |vtx2-vtx1|<=10\n";
    }
    if(graphClass==3){
        fGraph<<"c Layered Graph 20: all edges (vtx1,vtx2) satisfy |vtx2-vtx1|<=20\n";
    }
    if(graphClass==4){
        fGraph<<"c Layered Graph 10: all edges (vtx1,vtx2) satisfy |vtx2-vtx1|<=10\n";
    }
    fGraph<<"c Each edge appears once."<<endl;
    fGraph<<"c A line like 'b 50 13.7' indicate a demand of 13.7 for vtx 50. Source is 1"<<endl;
    fGraph<<"c A line starting with 'f' gives a graph label to be printed in stats"<<endl;
    fGraph<<"c If there is any line like 'd 3' that sets the bandwidth of each edge to 3 instead of default 1"<<endl;
    assert(labelGraph==NULL);
    labelGraph = new char[100];
    double bMaxDem = bMaxDemand[graphClass];
    int instanceID = (minId[graphClass]==INT_MIN?idThisInstance:idThisInstance-minId[graphClass]+1);
    char pad[10];
    strcpy(pad,"");
    if(bMaxDem<100)
        strcpy(pad,"0");
    if(bMaxDem<10)
        strcpy(pad,"00");
    //char padInst[10];
    //strcpy(padInst,"");
    //if(instanceID<100)
    //    strcpy(padInst,"0");
    //if(instanceID<10)
    //    strcpy(padInst,"00");
    //sprintf(labelGraph,"%s:%s%d-d%s%g",graphClassName,padInst,instanceID,pad, bMaxDemand[graphClass]);
    //no padding for instance id below
    sprintf(labelGraph,"%s&%d&%s%g",graphClassName,instanceID,pad, bMaxDemand[graphClass]);

    fGraph<<"l "<<labelGraph<<endl;
    if(bandwidth!=1)
        fGraph<<"d "<<bandwidth<<endl;
    fGraph<<"p edge "<<n<<" "<<m<<endl;

    for(int edg=0;edg<m;edg++){
          int i = ii[edg];                        //for(int i=n-1;i>=0;i--)
          int j = jj[edg];                        //    for(int j=n-1;j>=0;j--)
          assert((i!=j)&&(i<n)&&(i>=0)&&(j<n)); //        if(g[i][j]==1)
          fGraph<<"e "<<n-i<<" "<<n-j<<endl;//source is actually n-1, need reversing
    }

    for(int v=n-1;v>=0;v--)
        fGraph<<"b "<<n-v<<" "<<b[v]<<endl;           //reversing vtx
    fGraph<<"c edges end points:";
    for(int e = 0;e<m;e++)
        fGraph<<"("<<n-ii[e]<<","<<n-jj[e]<<"),";    //reversing
    fGraph.close();
    clog<<"Done.\n";
}
void loadGraph(char* filename){
    //C standard, sorry I recycled an existing old method
    FILE*in = fopen(filename,"r");
    if(in==NULL){
        cerr<<"Could not open input file "<<filename<<endl;
        exit(1);
    }
    char line [1000];
    g = NULL;
    int zeroIndexed = 0;
    int edgeId = 0;
    int demandsMentionned=0;
    int e1,e2;
    while(fgets(line, sizeof(line), in) != NULL){
        int linePos=0;
        while(line[linePos]==' ')
            linePos++;
        //clog<<"\nline[linePos]="<<line[linePos];
        switch(line[linePos]){
            case('c'):
                        break;//'c' stands for commentary in this file format
            case('l'):  //a graph label
                        assert(labelGraph==NULL);
                        labelGraph = new char[200];
                        sscanf(line+linePos+1,"%s",labelGraph);
                        clog<<"The graph label is "<<labelGraph<<endl;
                        break;
            case('d'):  
                        double bWidthLocal;
                        sscanf(line+linePos+1,"%lf",&bWidthLocal);
                        setBandwidth(bWidthLocal);
                        clog<<"Reading " <<filename<<" we set a bandwidth of "<<bandwidth<<endl;
            case('p'):      
                        while(line[linePos]<'0'||line[linePos]>'9') //jumping over word 'edge'
                            linePos++;                        
                        sscanf(line+ linePos,"%d %d",&n,&m);
                        clog<<"The input file states: n(vtx)="<<n<<", m(edg)="<<m<<endl;
                        initDataStructsLoadGraph();
                        b = new double[n];
                        for(int i=0;i<n;i++)
                            b[i] = 0;
                        break;
            case('e'):  
                        sscanf(line + linePos +1,"%d %d",&e1,&e2);
//                        cout<<e1<<"-"<<e2<<"   ";
                        if(e1==0||e2==0)
                            if(!zeroIndexed){
                                zeroIndexed = 1;
                                cerr<<"Attention: I consider the DIMACS input graph zero-indexed, as I found edges linking the 0 vtx. Not sure the results are ok because previous edges might have been considered 1-indexed.!!"<<endl;
                            }
                        e1+=zeroIndexed;
                        e2+=zeroIndexed;
                        if(e1>n||e2>n){
                            cerr<<"ERROR: I read vertices larger than "<<n<<" (declared vtx no) in your Dimacs file"<<endl;
                            cerr<<"Line was:"<<line<<endl;
                            exit(1);
                        }
                        if(g!=NULL){
                            g[n-e1][n-e2]=1;      //reversing
                            g[n-e2][n-e1]=1;
                            ii[edgeId] = n-e1;
                            jj[edgeId] = n-e2;
                            edgeId++;
                        }
                        else{
                            cerr<<"ERROR: I found edges in the graph before knowing the graph size (nr of vtx)"<<endl;
                            exit(1);
                        }                 
                        break;                        
            case('b'):  int vtx;
                        double bVal;
                        if(g==NULL){
                            cerr<<"ERROR: I found demand information (lines 'b ..') before knowing the graph size (nr of vtx)"<<endl;
                            exit(1);
                        }
                        sscanf(line + linePos +1,"%d %lf",&vtx,&bVal);
                        vtx+=zeroIndexed;
                        b[n-vtx ] = bVal;
                        demandsMentionned++;
                        break;
        }
    }            
    if(demandsMentionned<n){
        cerr<<"The inputfile"<<filename<<" does not mention "<<n<<" demands but only "<<demandsMentionned<<endl;
        cerr<<"Did you forget to provide the demands using lines like 'b 15 9.2'?. Can NOT continue\n";
        exit(1);
    }
    handleBandwidth();
}
char* getGraphLabelForStats(){
    if(labelGraph!=NULL)
        return labelGraph;
    labelGraph = new char[20];
    double bMaxDem = -1;
    for(int i=0;i<n;i++)
        bMaxDem = max(bMaxDem,b[i]);
    if(graphClass!=INT_MAX)
        if(bMaxDem!=bMaxDemand[graphClass]){
            cerr<<"Warning bMaxDemand for this graph class is "<<bMaxDemand[graphClass]
                <<" yet the actual maximum demand is "<<bMaxDem
                <<"\nIt is strange but I do not exit. Pay attention to filenames";
            exit(1);
        }
    char pad[10];
    strcpy(pad,"");
    if(bMaxDem<10)
        strcpy(pad,"00");
    if(bMaxDem<100)
        strcpy(pad,"0");
    sprintf(labelGraph,"unknownf-d%s%g",pad,bMaxDem);
    return labelGraph;
}

void printGraph(){
    for(int i=0;i<n;i++){
        for(int j=0;j<n;j++)
            clog<<g[i][j];
        clog<<endl;
    }
    for(int i=0;i<n;i++)
        cout<<b[i]<<",";
    cout<<endl;
}
void setBandwidth(double bb){
    clog<<"Setting the bandwidth to "<<bb<<endl;
    if((bandwidth>1)&&(bandwidth!=bb)){
        cerr<<"Attention: you set bandwidth to "<<bb<<" but it was set before to "<<bandwidth
            <<"\nPay attention if you indicated the bandwidth twice (command line and file)\n";
        cerr<<"Sorry, but I prefer to exit because the bandwidth value is not clear.\n";
        exit(1);
    }
    bandwidth = bb;
}
double getBandwidth(){
    return bandwidth;
}
void loadGraphAndDemands(int idGraph){
    idThisInstance = idGraph;
    graphClass = getClass(idGraph);
    strcpy(graphClassName,allGraphClassNames[graphClass]);
    setInstanceName(instanceName);
    srand(idGraph);
    char filename[200];
    sprintf(filename,"%s/%s.dat","instances",instanceName);
    
    ifstream input(filename);
    if(input.good()){
        clog<<"Reading graph and demands from "<<filename<<endl;
        loadGraph(filename);
        //sprintf(filename,"%s/%s.dat.copy","instances",instanceName);
        //saveGraphAndDemands(filename);
    }else{
        clog<<"Generating graphs and demands now\n";
        loadGraph();
        initDemands(bMaxDemand[graphClass]);  
        saveGraphAndDemands(filename);
        handleBandwidth();
    }
}
void loadGraphFromFile(char* fname){
    graphClass = INT_MAX;
    loadGraph(fname);
}
void genRndGraphAndDemandsOrLoadFromAssocFileWhenItExists(int rndSeed){
    if(bMaxDemandFromCmdLine>0)
        for(int i=0;i<KNOWNCLASSES;i++)
            bMaxDemand[i] = bMaxDemandFromCmdLine;
    loadGraphAndDemands(rndSeed);
}
