/*---------------+---------------------------------------------------------------+--------------+
                 | Author: Daniel Porumbel 2018       daniel.porumbel@cnam.fr    |
                 |License: Creative Commons Attribution 3.0 Unported License     |
                 |         http://creativecommons.org/licenses/by/3.0/           |             
                 +--------------------------------------------------------------*/
#include <iostream>
#include <cmath>
#include <climits>
#include <iomanip>
#include <locale>
#include <sstream>
#include <string.h>
#include "statManager.h"
using namespace std;

void statManager::setLenOutput(int x){
        lenOutput = x;
}
void statManager::init(){
        idCurrVal = 0;
        vals = new double[maxVals];
}
statManager::statManager(){
        lenOutput = 5;
        maxVals = 200;
        init();
}
statManager::statManager(int len){
        lenOutput = len;
        init();
}
statManager::statManager(int len, int x){
        lenOutput = len;
        maxVals = x;
        init();
}
void statManager::addValue(double d){
        vals[idCurrVal] = d;
        idCurrVal++;
        if(idCurrVal==maxVals){
            cerr<<"Increase the number of vals (maxVals) usable by the statManager\n";
            cerr<<"statManager: Can not add further values\n";
            idCurrVal--;
        }
}
void statManager::addValue(int x){
        addValue((double)x);
}
int statManager::getPrecision(double number){
    int precision = 1;
    if(number>=1)
        precision = 2;
    if(number>=10)
        precision = 3;
    if(number>=1000)
        precision = 4;
    if(number>=10000)
        precision = 5;
    return precision;
}
string statManager::getString(double number){
    int prec = getPrecision(abs(number));
    std::ostringstream sBuilder;   
    if((number>=0.001)||(number==0))
        sBuilder <<left<<setw(lenOutput)<<setprecision(prec)<<number;
    else
        sBuilder <<left<<setw(lenOutput)<<setprecision(prec)<<"<1e-3";
    //char* test = (char*)sBuilder.str().c_str();
    return sBuilder.str();
}
void statManager::calc_minmax_avg_dev(){
    double n = (double)idCurrVal;//use double to be sure on some precision calculations
    if(n<=0){
        cerr<<"statManager: you seem to want stat when you did not call addValue(..)\n";
        return;
    }
    min = vals[0];
    max = vals[0];
    avg = vals[0];
    for(int i=1;i<n;i++){
        if(vals[i]>max)
            max = vals[i];
        if(vals[i]<min)
            min = vals[i];
        avg+=vals[i];
    }
    avg = avg/n;
    dev = 0;
    for(int i=0;i<n;i++)
        dev += (vals[i]-avg)*(vals[i]-avg);
    dev = dev / n;
    dev = sqrt(dev);
}
char* statManager::getStats(){
    calc_minmax_avg_dev();
    std::ostringstream sBuilder;   
    sBuilder <<getString(avg);
    sBuilder <<"&"<<getString(dev);
    sBuilder <<"&"<<getString(min);
    sBuilder <<"M"<<getString(max);
    char* test = strdup(sBuilder.str().c_str());
    return test;
}
double statManager::getAvg(){
    calc_minmax_avg_dev();
    return avg;
}
int statManager::getNrValues(){
    return idCurrVal;
}

statManager::~statManager(){
    delete[] vals;
}
